import pandas as pd
import numpy as np
from datetime import datetime
import opendp.prelude as dp
import random

dp.enable_features("contrib", "floating-point", "honest-but-curious")



def dataframe_domain(public_key_sets=None):
    """Creates a domain representing the set of all data frames.
    
    Assumes column names and types are public information.
    Key sets optionally named for columns in `public_key_sets` are considered public information.

    Two data frames differing in their public information 
    are considered to have a data set distance of infinity.
    """
    return dp.user_domain(
        "DataFrameDomain", lambda x: isinstance(x, pd.DataFrame), public_key_sets
    )


def series_domain():
    """Creates a domain representing the set of all series.

    Assumes series name and type are public information.

    Two series differing in their public information 
    are considered to have a data set distance of infinity.
    """
    return dp.user_domain("SeriesDomain", lambda x: isinstance(x, pd.Series))

def identifier_distance():
    """Symmetric distance between the id sets."""
    return dp.user_distance("IdentifierDistance")


def approx_concentrated_divergence():
    """symmetric distance between the id sets"""
    return dp.user_distance("ApproxConcentratedDivergence()")

def make_preprocess_location():
    """Create a 1-stable transformation to bin `merch_postal_code` by city"""

    def categorize_city(code):
        if code.startswith("5"):
            return "Medellin"
        elif code.startswith("11"):
            return "Bogota"
        elif code.startswith("70"):
            return "Brasilia"
        else:
            return "Santiago"

    def location_preprocess(df):
        loc_df = df.copy()
        # Convert merchant_postal_code into str type
        loc_df["merch_postal_code"] = loc_df["merch_postal_code"].astype(str)
        # Apply the function to create a new column
        loc_df["city"] = loc_df["merch_postal_code"].apply(
            categorize_city
        )
        return loc_df

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=identifier_distance(),
        output_domain=dataframe_domain(),
        output_metric=identifier_distance(),
        function=location_preprocess,
        stability_map=lambda d_in: d_in,
    )

def make_truncate_time(start_date, end_date, time_col):
    """Create a transformation that filters the data to a given time frame.
    
    WARNING: Assumes that the data has at most one contribution per individual per week.
    """
    number_of_timesteps = (end_date - start_date).days // 7

    def time_preprocess(df):
        df = df.copy()

        # Convert time_col into datetime type
        df[time_col] = pd.to_datetime(df[time_col])

        # Filter the DataFrame based on the specified dates
        return df[(df[time_col] >= start_date) & (df[time_col] <= end_date)]

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=identifier_distance(),
        output_domain=dataframe_domain(),
        output_metric=dp.symmetric_distance(),
        function=time_preprocess,
        stability_map=lambda d_in: d_in * number_of_timesteps,
    )

def make_sum_by(column, by, bounds):
    """Create a transformation that computes the grouped bounded sum of `column`"""
    L, U = bounds
    def function(df):
        df = df.copy()
        df[column] = df[column].clip(*bounds)
        return df.groupby(by)[column].sum()

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=dp.symmetric_distance(),
        output_domain=series_domain(),
        output_metric=dp.l2_distance(T=float),
        function=function,
        stability_map=lambda d_in: np.sqrt(d_in) * max(abs(L), U),
    )


def make_private_sum_by(column, by, bounds, scale):
    """Create a measurement that computes the grouped bounded sum of `column`"""
    space = dp.vector_domain(dp.atom_domain(T=int)), dp.l2_distance(T=float)
    m_gauss = space >> dp.m.then_gaussian(scale)
    t_sum = make_sum_by(column, by, bounds)

    def function(df):
        exact = t_sum(df)
        # print(exact)
        noisy_sum = pd.Series(
            np.maximum(m_gauss(exact.to_numpy().flatten()), 0), 
        )
        # print(noisy_sum)
        noisy_sum=noisy_sum.to_frame(name=column)
        noisy_sum[by] = exact.index
        return noisy_sum

    return dp.m.make_user_measurement(
        input_domain=dataframe_domain(public_key_sets=[by]),
        input_metric=dp.symmetric_distance(),
        output_measure=dp.zero_concentrated_divergence(T=float),
        function=function,
        privacy_map=lambda d_in: m_gauss.map(t_sum.map(d_in)),
    )

def make_filter(column,entry):
        """filters offline entries"""
        
        def function(df):
            df = df.copy()
            return df[(df[column] == entry)]


        return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=identifier_distance(),
        output_domain=dataframe_domain(),
        output_metric=identifier_distance(),
        function=function,
        stability_map=lambda d_in: d_in,
    )

def make_preprocess_merchant():
    """Create a 1-stable transformation to bin `merch_postal_code` by city"""

    def categorize_merchant(merch):
        if merch in ['Hotels/Motels','Restaurants','Bars/Discotheques']:
            return "luxury"
        elif merch in ['Grocery Stores/Supermarkets','Drug Stores/Pharmacies','General Retail Stores','Utilities: Electric, Gas, Water','Hospitals']:
            return "essential"
        else:
            return "other"

    def merchant_preprocess(df):
        loc_df = df.copy()
        # Convert merchant_postal_code into str type
        loc_df["merch_category"] = loc_df["merch_category"].astype(str)
        # Apply the function to create a new column
        loc_df["merch_super_category"] = loc_df["merch_category"].apply(
            categorize_merchant
        )
        return loc_df

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=identifier_distance(),
        output_domain=dataframe_domain(),
        output_metric=identifier_distance(),
        function=merchant_preprocess,
        stability_map=lambda d_in: d_in,
    )

def get_coordinates(df:pd.DataFrame):
    postal_codes=df["merch_postal_code"].unique()


    postal_code = {
        'Medellin': [ code  for code in postal_codes if str(code).startswith('5')],
        'Bogota': [ code  for code in postal_codes if str(code).startswith('11')],
        'Brasilia': [ code  for code in postal_codes if str(code).startswith('70')],
        'Santiago': [ code  for code in postal_codes if not str(code).startswith('5') and not str(code).startswith('11') and not str(code).startswith('70')]
    }

    # Reference coordinates
    reference_coords = {
        "Medellin": (6.2476, -75.5658),
        "Bogota": (4.7110, -74.0721),
        "Brasilia": (-15.7975, -47.8919),
        "Santiago": (-33.4489, -70.6693)
    }

    # Function to generate unique coordinates
    def generate_unique_coords(base_lat, base_lon, num_coords):
        coords = []
        for _ in range(num_coords):
            # Slightly vary the base coordinates
            lat_variation = random.uniform(-1.5, +1.5)
            lon_variation = random.uniform(-1.5, +1.5)
            new_lat = base_lat + lat_variation
            new_lon = base_lon + lon_variation
            coords.append((new_lat, new_lon))
        return coords

    # Assign unique coordinates to each postal code
    postal_code_coords = {}
    for segment, codes in postal_code.items():
        base_lat, base_lon = reference_coords[segment]
        unique_coords = generate_unique_coords(base_lat, base_lon, len(codes))
        for code, coord in zip(codes, unique_coords):
            postal_code_coords[code] = coord

    df['Latitude'], df['Longitude'] = zip(*df['merch_postal_code'].map(postal_code_coords))
    return df


def make_filter_rows(column_name, value):
    """Create a transformation that filters the rows based on the value of a column `column_name`.

    """

    def filter_rows(df):
        df = df.copy()

        # Filter the rows
        return df[df[column_name] == value]

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=dp.symmetric_distance(),
        output_domain=dataframe_domain(),
        output_metric=dp.symmetric_distance(),
        function=filter_rows,
        stability_map=lambda d_in: d_in,
    )

def make_filter_rows_with_country(column_name, country_code_prefix):
    """Create a transformation that filters the rows based on the value of a column `column_name`.
    """
    def filter_rows(df):
        df = df.copy()

        # Filter the rows
        return df[df["merch_postal_code"].astype(str).str.startswith(country_code_prefix)]

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=dp.symmetric_distance(),
        output_domain=dataframe_domain(),
        output_metric=dp.symmetric_distance(),
        function=filter_rows,
        stability_map=lambda d_in: d_in,
    )

def make_private_nb_transactions_avg_count(merch_category, upper_bound , dp_dataset_size, scale = 1.0):
    """Create a measurement that computes the grouped bounded sum"""

    def compute_private_sum(df):
        df = df.copy()
        sum = df[df["merch_category"]==merch_category]["nb_transactions"].clip(lower=0, upper=upper_bound).sum()
        dp_sum = np.random.laplace(loc=sum, scale=scale)
        return dp_sum/dp_dataset_size

    return dp.m.make_user_measurement(
        input_domain=dataframe_domain(),
        input_metric=dp.symmetric_distance(),
        output_measure=dp.max_divergence(T=float),
        function=compute_private_sum,
        privacy_map=lambda d_in: d_in*upper_bound,
    )

def make_private_count(city_col, city, epsilon):
    """Create a measurement that computes the count of the unique zip codes in the given city"""

    def private_count(df):
        df = df.copy()
        zip_code_list = df[df[city_col]==city]["merch_postal_code"].unique()
        # (where TIA stands for Atomic Input Type)
        input_space = dp.vector_domain(dp.atom_domain(T=float)), dp.symmetric_distance()
        # (where TIA stands for Atomic Input Type)
        # TODO: what should be the sensitivity here? Many merchant records can contribute to the same zip code
        # So, scale = 1, is a good enough choice.
        count_meas = input_space >> dp.t.then_count() >> dp.m.then_laplace(1.)
        dp_count = count_meas(zip_code_list)
        return dp_count
    
    return dp.m.make_user_measurement(
        input_domain=dataframe_domain(),
        input_metric=dp.symmetric_distance(),
        output_measure=dp.max_divergence(T=int),
        function=private_count,
        privacy_map=lambda d_in: d_in*epsilon,
    )

def make_preprocess_merchant_mobility():
    """Create a 1-stable transformation to bin `merch_postal_code` by city"""

    def categorize_merchant(merch):
        if merch in ['General Retail Stores','Restaurants','Bars/Discotheques']:
            return "retail_and_recreation"
        elif merch in ['Grocery Stores/Supermarkets','Drug Stores/Pharmacies']:
            return "grocery_and_pharmacy"
        elif merch in ['Airlines']:
            return "transit_stations"
        else:
            return "other"

    def merchant_preprocess(df):
        loc_df = df.copy()
        # Convert merchant_postal_code into str type
        loc_df["merch_category"] = loc_df["merch_category"].astype(str)
        # Apply the function to create a new column
        loc_df["merch_super_category"] = loc_df["merch_category"].apply(
            categorize_merchant
        )
        return loc_df

    return dp.t.make_user_transformation(
        input_domain=dataframe_domain(),
        input_metric=identifier_distance(),
        output_domain=dataframe_domain(),
        output_metric=identifier_distance(),
        function=merchant_preprocess,
        stability_map=lambda d_in: d_in,
    )

def preprocess_google_mobility(df:pd.DataFrame,start_date:datetime, end_date:datetime,  city:str, category:str, offset: datetime=None):

    # df = df.copy()

    def region_filter(df, country_code, region_1, region_2, all=False):
        df=df.copy()
        mask = ((df["country_region_code"].isin(country_code)) & (df["sub_region_1"].isin(region_1)))
        if all:
            mask = (mask & (df["sub_region_2"].isin(region_2)))
        return df[mask]
    
    def time_preprocess(df):
        df = df.copy()

        # Filter the DataFrame based on the specified dates
        return df[(df["date"] >= start_date) & (df["date"] <= end_date)]
    
    if city == "Bogota":
        df = region_filter(df, ["CO"], ["Bogota"], "")
    elif city == "Medellin":
        df = region_filter(df, ["CO"], ["Antioquia"], ["Medellin"], all=True)
    elif city == "Santiago":
        df = region_filter(df, ["CL"], ["Santiago Metropolitan Region"], ["Santiago Province"], all=True)
    elif city == "Brasilia":
        df = region_filter(df, ["BR"], ["Federal District"], "")
    else:
        raise ValueError("Invalid city")

    df = df.drop(["place_id", "country_region_code", "country_region", "sub_region_1", "sub_region_2", "metro_area", "iso_3166_2_code", "census_fips_code","parks_percent_change_from_baseline","workplaces_percent_change_from_baseline","residential_percent_change_from_baseline"], axis=1)
    
    # Ensure the date column is in datetime format
    df['date'] = pd.to_datetime(df['date'])
    
    df=time_preprocess(df)
    # Set the date column as the index
    df.set_index('date', inplace=True)

    # Group by week and calculate the sum
    df_weekly = df.resample('W', label="right").sum()

    # Shift dates back to original dates
    if offset:
        
        if not (offset >= (df_weekly.index[0]-pd.DateOffset(7)) and  offset<=df_weekly.index[1]):
            raise ValueError("Invalid offset date")
        
        df_weekly.index = df_weekly.index - pd.DateOffset(days=(df_weekly.index[0] - offset).days)
   
    # Reset the index to keep the first date of each week
    df_weekly.reset_index(inplace=True)
    
    df_weekly.rename(columns={"retail_and_recreation_percent_change_from_baseline": "retail_and_recreation", "grocery_and_pharmacy_percent_change_from_baseline": "grocery_and_pharmacy", "transit_stations_percent_change_from_baseline": "transit_stations"}, inplace=True)
    
    # Filter the dataframe based on the category
    df_final = df_weekly[["date", category]]
    
    return df_final



import dash
from dash import dcc, html
from dash.dependencies import Input, Output
from datetime import datetime
import plotly.express as px
import pandas as pd
import numpy as np
import sys
import os
import random
from datetime import datetime
import opendp.prelude as dp

dp.enable_features("contrib", "floating-point", "honest-but-curious")

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from DP_epidemiology.utilities import *

from DP_epidemiology.hotspot_analyzer import hotspot_analyzer 
from DP_epidemiology.mobility_analyzer import mobility_analyzer
from DP_epidemiology.pandemic_adherence_analyzer import pandemic_adherence_analyzer
from DP_epidemiology.contact_matrix import get_age_group_count_map, get_contact_matrix, get_pearson_similarity


def create_hotspot_dash_app(df:pd.DataFrame):
    cities = {
    "Medellin": (6.2476, -75.5658),
    "Bogota": (4.7110, -74.0721),
    "Brasilia": (-15.7975, -47.8919),
    "Santiago": (-33.4489, -70.6693)
    }
    app = dash.Dash(__name__)

    app.layout = html.Div([
        dcc.DatePickerSingle(
            id='start-date-picker',
            date='2019-01-01'
        ),
        dcc.DatePickerSingle(
            id='end-date-picker',
            date='2019-12-31'
        ),
        dcc.Slider(
            id='epsilon-slider',
            min=0,
            max=10,
            step=0.1,
            value=1,
            marks={i: str(i) for i in range(11)}
        ),
        dcc.Dropdown(
            id='city-dropdown',
            options=[{'label': city, 'value': city} for city in cities.keys()],
            value='Medellin'
        ),
        dcc.Graph(id='geo-plot')
    ])

    @app.callback(
        Output('geo-plot', 'figure'),
        [Input('start-date-picker', 'date'),
         Input('end-date-picker', 'date'),
         Input('epsilon-slider', 'value'),
         Input('city-dropdown', 'value')]
    )
    def update_graph(start_date, end_date, epsilon, city):
        start_date = datetime.strptime(start_date, '%Y-%m-%d')
        end_date = datetime.strptime(end_date, '%Y-%m-%d')
        
        # Filter data using hotspot_analyser
        output = hotspot_analyzer(df, start_date, end_date, city, epsilon)
        filtered_df = get_coordinates(output)

        # Plot using Plotly Express
        fig = px.scatter_geo(
            filtered_df,
            lat='Latitude',
            lon='Longitude',
            color='nb_transactions',
            size='nb_transactions',
            hover_name='merch_postal_code',
            hover_data={'merch_postal_code': True, 'nb_transactions': True, 'Latitude': False, 'Longitude': False},
            projection='mercator',
            title=f"Transaction Locations in {city} from {start_date.date()} to {end_date.date()} with epsilon={epsilon}",
            color_continuous_scale=px.colors.sequential.Plasma
        )

        # Center the map around the selected city
        fig.update_geos(
            center=dict(lat=cities[city][0], lon=cities[city][1]),
            projection_scale=2.5  # Zoom level
        )

        return fig

    return app

def create_mobility_dash_app(df:pd.DataFrame):
    cities = {
        "Medellin": (6.2476, -75.5658),
        "Bogota": (4.7110, -74.0721),
        "Brasilia": (-15.7975, -47.8919),
        "Santiago": (-33.4489, -70.6693)
        }
    app = dash.Dash(__name__)
    category_list = ['grocery_and_pharmacy', 'transit_stations', 'retail_and_recreation',"other"]
    app.layout = html.Div([
            dcc.DatePickerSingle(
                id='start-date-picker',
                date='2019-01-01'
            ),
            dcc.DatePickerSingle(
                id='end-date-picker',
                date='2019-12-31'
            ),
            dcc.Slider(
                id='epsilon-slider',
                min=0,
                max=10,
                step=0.1,
                value=1,
                marks={i: str(i) for i in range(11)}
            ),
            dcc.Dropdown(
                id='city-dropdown',
                options=[{'label': city, 'value': city} for city in cities.keys()],
                value='Medellin'
            ),
            dcc.Dropdown(
                id='category-list-dropdown',
                options=[{'label': category, 'value': category} for category in category_list],
                value='transit_stations'
            ),
            dcc.Graph(id='mobility-graph')
        ])

    # Callback to update the graph based on input values
    @app.callback(
        Output('mobility-graph', 'figure'),
        [Input('start-date-picker', 'date'),
        Input('end-date-picker', 'date'),
        Input('city-dropdown', 'value'),
        Input('category-list-droopdown', 'value'),
        Input('epsilon-slider', 'value')]
    )
    def update_graph(start_date, end_date, city_filter,category, epsilon):
        # Convert date strings to datetime objects
        start_date = datetime.strptime(start_date, '%Y-%m-%d')
        end_date = datetime.strptime(end_date, '%Y-%m-%d')

        # Call the mobility_analyser function
        filtered_df = mobility_analyzer(df, start_date, end_date, city_filter, category, epsilon)

        # Plot using Plotly Express
        fig = px.line(
            filtered_df,
            x='date',
            y='nb_transactions',
            title=f"Mobility Analysis for {city_filter} and category{category} from {start_date.date()} to {end_date.date()} with epsilon={epsilon}",
            labels={'nb_transactions': 'Number of Transactions', 'date': 'Date'}
        )

        return fig
    return app

def create_pandemic_adherence_dash_app(df:pd.DataFrame):
    cities = {
        "Medellin": (6.2476, -75.5658),
        "Bogota": (4.7110, -74.0721),
        "Brasilia": (-15.7975, -47.8919),
        "Santiago": (-33.4489, -70.6693)
        }
    entry_types=["luxury","essential","other"]
    app = dash.Dash(__name__)

    app.layout = html.Div([
            dcc.DatePickerSingle(
                id='start-date-picker',
                date='2019-01-01'
            ),
            dcc.DatePickerSingle(
                id='end-date-picker',
                date='2019-12-31'
            ),
            dcc.Slider(
                id='epsilon-slider',
                min=0,
                max=10,
                step=0.1,
                value=1,
                marks={i: str(i) for i in range(11)}
            ),
            dcc.Dropdown(
                id='city-dropdown',
                options=[{'label': city, 'value': city} for city in cities.keys()],
                value='Medellin'
            ),
            dcc.Dropdown(
                id='entry-type-dropdown',
                options=[{'label': entry_type, 'value': entry_type} for entry_type in entry_types],
                value='luxury'
            ),
            dcc.Graph(id='pandemic-adherence-graph')
        ])

    # Callback to update the graph based on input values
    @app.callback(
        Output('pandemic-adherence-graph', 'figure'),
        [Input('start-date-picker', 'date'),
        Input('end-date-picker', 'date'),
        Input('city-dropdown', 'value'),
        Input('entry-type-dropdown', 'value'),
        Input('epsilon-slider', 'value')]
    )
    def update_graph(start_date, end_date, city_filter,essential_or_luxury, epsilon):
        # Convert date strings to datetime objects
        start_date = datetime.strptime(start_date, '%Y-%m-%d')
        end_date = datetime.strptime(end_date, '%Y-%m-%d')

        # Call the mobility_analyser function
        filtered_df = pandemic_adherence_analyzer(df, start_date, end_date, city_filter,essential_or_luxury, epsilon)

        # Plot using Plotly Express
        fig = px.line(
            filtered_df,
            x='date',
            y='nb_transactions',
            title=f"Pandemic Stage Analysis for {city_filter} from {start_date.date()} to {end_date.date()} with epsilon={epsilon}",
            labels={'nb_transactions': 'Number of Transactions', 'date': 'Date'}
        )

        return fig
    return app

def create_contact_matrix_dash_app(df:pd.DataFrame):
    cities = {
        "Medellin": (6.2476, -75.5658),
        "Bogota": (4.7110, -74.0721),
        "Brasilia": (-15.7975, -47.8919),
        "Santiago": (-33.4489, -70.6693)
    }
    # Initialize the Dash app
    app = dash.Dash(__name__)

    app.layout = html.Div([
            dcc.DatePickerSingle(
                id='start-date-picker',
                date='2019-01-01'
            ),
            dcc.DatePickerSingle(
                id='end-date-picker',
                date='2019-12-31'
            ),
            dcc.Slider(
                id='epsilon-slider',
                min=0,
                max=10,
                step=0.1,
                value=1,
                marks={i: str(i) for i in range(11)}
            ),
            dcc.Dropdown(
                id='city-dropdown',
                options=[{'label': city, 'value': city} for city in cities.keys()],
                value='Medellin'
            ),
            dcc.Graph(id='matrix-heatmap'),

        # Output table to display contact matrix values
        html.Div(id='matrix-output', style={'whiteSpace': 'pre-line'})
    ])

    # Callback to update the graph and output based on inputs
    @app.callback(
        [Output('matrix-heatmap', 'figure'),
        Output('matrix-output', 'children')],
        [Input('start-date-picker', 'date'),
        Input('end-date-picker', 'date'),
        Input('city-dropdown', 'value'),
        Input('epsilon-slider', 'value')]
    )
    def update_contact_matrix(start_date, end_date, city, epsilon):
        start_date = datetime.strptime(start_date, '%Y-%m-%d')
        end_date = datetime.strptime(end_date, '%Y-%m-%d')

        # Get age group count map
        age_group_count_map = get_age_group_count_map(df, start_date, end_date, city, epsilon)
        age_group_sample_size = list(age_group_count_map.values())
        
        # Hardcoded population distribution for the example
        age_group_population_distribution = [8231200, 7334319, 6100177]

        # Generate contact matrix
        contact_matrix = get_contact_matrix(age_group_sample_size, age_group_population_distribution)

        # Generate a heatmap for the contact matrix
        fig = px.imshow(contact_matrix,
                        labels=dict(x="Age Group", y="Age Group", color="Contact Rate"),
                        x=['20-30', '30-40', '50-60'],
                        y=['20-30', '30-40', '50-60'],
                        title=f"contact matrix for {city} from {start_date.date()} to {end_date.date()} with epsilon={epsilon} and pearson similarity={get_pearson_similarity(contact_matrix)}",
                        zmin=2.5, zmax=3.5,
                        color_continuous_scale='Blues')

        # Convert the matrix to a readable format
        matrix_output = f"Contact Matrix:\n{np.array_str(contact_matrix, precision=2, suppress_small=True)}"
        
        return fig, matrix_output
    
    return app


import pandas as pd
import numpy as np
import sys
import os
from datetime import datetime
import scipy.stats as stats
import opendp.prelude as dp

dp.enable_features("contrib", "floating-point", "honest-but-curious")

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from DP_epidemiology.utilities import *

def mobility_analyzer_airline(df:pd.DataFrame,start_date:datetime,end_date:datetime,city: str, epsilon:float):
    """final function to predict hotspots"""
    bounds = (0, 600)
    upper_bound=600
    transaction_data_col = "nb_transactions"
    groupby_col = "date"

    city_col="city"
    time_col="date"
    merch_category_col="merch_category"
    merch_filter="Airlines"
    

    """time steps calculation"""
    nb_timesteps = (end_date - start_date).days // 7

    """scale calculation"""
    scale=(np.sqrt(3.0)*nb_timesteps*upper_bound)/epsilon

    new_df=df.copy()


    analyzer=(
    make_preprocess_location()
    >>make_filter(city_col,city)
    >>make_filter(merch_category_col,merch_filter)
    >>make_truncate_time(start_date, end_date, time_col)
    >>make_private_sum_by(transaction_data_col, groupby_col, bounds, scale)
   )

    return analyzer(new_df)

def mobility_analyzer(df:pd.DataFrame,start_date:datetime,end_date:datetime,city: str,category:str, epsilon:float):
    """final function to predict hotspots"""
    bounds = (0, 600)
    upper_bound=600
    transaction_data_col = "nb_transactions"
    groupby_col = "date"

    city_col="city"
    time_col="date"
    merch_category_col="merch_super_category"
    # merch_filter="Airlines"
    

    """time steps calculation"""
    nb_timesteps = (end_date - start_date).days // 7

    """scale calculation"""
    scale=(np.sqrt(3.0)*nb_timesteps*upper_bound)/epsilon

    new_df=df.copy()


    analyzer=(
    make_preprocess_location()
    >>make_preprocess_merchant_mobility()
    >>make_filter(city_col,city)
    >>make_filter(merch_category_col, category)
    >>make_truncate_time(start_date, end_date, time_col)
    >>make_private_sum_by(transaction_data_col, groupby_col, bounds, scale)
   )

    return analyzer(new_df)

def mobility_validation_with_google_mobility(df_transactional_data:pd.DataFrame, df_google_mobility_data:pd.DataFrame, start_date:datetime, end_date:datetime, city:str, category:str, epsilon:float):
    df_transactional_mobility= mobility_analyzer(df_transactional_data,start_date,end_date,city,category,epsilon)
    offset=df_transactional_mobility["date"][0]
    df_google_mobility = preprocess_google_mobility(df_google_mobility_data,start_date,end_date,city,category,offset)

    length =min(len(df_transactional_mobility),len(df_google_mobility))

    r, p = stats.pearsonr(df_transactional_mobility[category][:length], df_google_mobility[category][:length])
    print(f"Scipy computed Pearson r: {r} and p-value: {p}")